import tempfile
import transaction
from Acquisition import aq_base
from zope.interface import directlyProvides, directlyProvidedBy
from zope.event import notify
from zope.component import getUtility
from zope.component.interfaces import ComponentLookupError
from zope.app.intid.interfaces import IIntIds
from zope.location.pickling import locationCopy
from plone.relations.interfaces import IComplexRelationshipContainer
from plone.app.relations.interfaces import IRetainOnCopy
from plone.app.relations.interfaces import ICopyPendingAdd
from plone.app.relations.interfaces import IRelationshipSource
from plone.app.relations.interfaces import IRelationshipTarget
from plone.app.relations.interfaces import RelationSourceDeleted
from plone.app.relations.interfaces import RelationTargetDeleted
from plone.app.relations.interfaces import HoldingRelationError
from zc.relationship.interfaces import IBidirectionalRelationshipIndex


def handleSourceTargetDelete(ob, event):
    """Fires relevant events for all references related to the object"""
    try:
        source = IRelationshipSource(ob)
        target = IRelationshipTarget(ob)
    except (ComponentLookupError, TypeError):
        return
    # we tuplify the generator to prevent errors if we need to delete elements
    rels = tuple(source.getRelationships())
    for rel in rels:
        notify(RelationSourceDeleted(rel, ob))
    rels = tuple(target.getRelationships())
    for rel in rels:
        notify(RelationTargetDeleted(rel, ob))

def deleteReferenceOnSourceDelete(rel, event):
    """deleted references related to the object"""
    index = (IBidirectionalRelationshipIndex.providedBy(rel.__parent__) and
             rel.__parent__ or None)
    if len(rel.sources) > 1:
        rel.sources.remove(event.source)
        if index is not None:
            index.reindex(rel)
    elif event.source in rel.sources and index is not None:
        index.remove(rel)

def deleteReferenceOnTargetDelete(rel, event):
    """deleted references related to the object"""
    index = (IBidirectionalRelationshipIndex.providedBy(rel.__parent__) and
             rel.__parent__ or None)
    if len(rel.targets) > 1:
        rel.targets.remove(event.target)
        if index is not None:
            index.reindex(rel)
    elif event.target in rel.targets and index is not None:
        index.remove(rel)

def raiseHoldingExceptionOnTargetDelete(rel, event):
    """raises an exception when the target of an existing IHoldingRelation
    is deleted"""
    raise HoldingRelationError(event.target, rel)

def markCopyOnCopy(obj, event):
    obj.__orig_object__ = event.original
    directlyProvides(obj, ICopyPendingAdd)

def copyRelationsOnSourceCopy(obj, event):
    """Copies all source relaitonships marked with IRetainOnCopy
    when the source object is copied"""
    orig_obj = obj.__orig_object__.aq_inner
    source = IRelationshipSource(orig_obj)
    cur_ifaces = directlyProvidedBy(orig_obj)
    directlyProvides(obj, *[i for i in cur_ifaces if i is not ICopyPendingAdd])
    delattr(obj, '__orig_object__')
    copy_filter = IRetainOnCopy.providedBy
    # this is not efficient to source objects with a huge number of
    # relations
    rels = source.getRelationships(rel_filter=copy_filter)
    container = getUtility(IComplexRelationshipContainer, name='relations')
    unwrapped_obj = aq_base(obj)
    for rel in rels:
        # copy the relationship
        rel_copy = locationCopy(aq_base(rel))
        # The references on the copy becames copies themselves we need
        # to make sure all such references point to the originals:
        rel_copy.__dict__ = rel.__dict__
        # We should also remove existing ILocation pointers
        rel_copy.__parent__ = rel_copy.__name__ = None
        # Now we add the relationship (with wrong sources for now) to
        # give it a context.
        container.add(rel_copy)
        # replace the old sources with just the copy
        rel_copy.sources = (obj,)




# A tiny monkey patch due to some re-organization of future BTree modules
try:
    from BTrees.OOBTree import BTree
except ImportError:
    import BTrees.OOBTree
    import BTrees.IOBTree
    import BTrees.OIBTree
    import BTrees.IIBTree
    import BTrees.IFBTree
    BTrees.OOBTree.BTree = BTrees.OOBTree.OOBTree
    BTrees.OOBTree.Set = BTrees.OOBTree.OOSet
    BTrees.OOBTree.Bucket = BTrees.OOBTree.OOBucket
    BTrees.OOBTree.TreeSet = BTrees.OOBTree.OOTreeSet
    BTrees.IOBTree.BTree = BTrees.IOBTree.IOBTree
    BTrees.IOBTree.Set = BTrees.IOBTree.IOSet
    BTrees.IOBTree.Bucket = BTrees.IOBTree.IOBucket
    BTrees.IOBTree.TreeSet = BTrees.IOBTree.IOTreeSet
    BTrees.OIBTree.BTree = BTrees.OIBTree.OIBTree
    BTrees.OIBTree.Set = BTrees.OIBTree.OISet
    BTrees.OIBTree.Bucket = BTrees.OIBTree.OIBucket
    BTrees.OIBTree.TreeSet = BTrees.OIBTree.OITreeSet
    BTrees.IIBTree.BTree = BTrees.IIBTree.IIBTree
    BTrees.IIBTree.Set = BTrees.IIBTree.IISet
    BTrees.IIBTree.Bucket = BTrees.IIBTree.IIBucket
    BTrees.IIBTree.TreeSet = BTrees.IIBTree.IITreeSet
    BTrees.IFBTree.BTree = BTrees.IFBTree.IFBTree
    BTrees.IFBTree.Set = BTrees.IFBTree.IFSet
    BTrees.IFBTree.Bucket = BTrees.IFBTree.IFBucket
    BTrees.IFBTree.TreeSet = BTrees.IFBTree.IFTreeSet



# Some utilities for mail
from email.Message import Message
from email import Header
from email.MIMEText import MIMEText
from email.MIMEMultipart import MIMEMultipart
from email.Charset import add_charset, QP, SHORTEST
from StructuredText.StructuredText import HTML

# Add a charset for utf8 that should be encoded quopri,
# e.g. we assume mostly latin chars, if not the encoding will
# be very space inefficient
add_charset('utf8', SHORTEST, QP, 'utf8')

def decode_header(value):
    """Converts an encoded header into a unicode string::

        >>> decode_header(
        ...       'Je les =?utf-8?b?ZMODwql0w4PCqHN0ZQ==?= oui?')
        u'Je les d\\xc3\\xa9t\\xc3\\xa8ste oui?'

    """
    encoded_strings = Header.decode_header(value)
    header_val = Header.Header()
    for string, encoding in encoded_strings:
        header_val.append(string, encoding, errors='replace')
    return unicode(header_val)

def construct_simple_encoded_message(from_addr, to_addr, subject, body,
                                  other_headers=None, encoding='utf8'):
    """All inputs to this method are expected to be unicode or ascii.

    We should be able to pass in some arbitrary unicode stuff and get back
    a sensible encoded message:

        >>> m = construct_simple_encoded_message(u'test@example.com',
        ...     u'test@example.com',
        ...     u'Un Subj\xc3\xa9t',
        ...     u'A simple body with some non ascii t\xc3\xa9xt',
        ...     other_headers = {'X-Test': u't\xc3\xa9st'})
        >>> print m.as_string()
        From: test@example.com
        To: test@example.com
        Subject: Un =?utf8?b?U3ViasODwql0?=
        X-Test: =?utf8?b?dMODwqlzdA==?=
        MIME-Version: 1.0
        Content-Type: text/plain; charset="utf8"
        Content-Transfer-Encoding: quoted-printable
        <BLANKLINE>
        A simple body with some non ascii t=C3=83=C2=A9xt
    """
    if other_headers is None:
        other_headers = {}
    m = Message()
    # If from is not in the set of other headers use the from addr
    # otherwise we may be masquerading as someone else
    if 'From' not in other_headers:
        m['From'] = encode_header(from_addr, encoding)
    m['To'] = encode_header(to_addr, encoding)
    m['Subject'] = encode_header(subject, encoding)
    for key, val in other_headers.items():
        m[key] = encode_header(val, encoding)
    body = body.encode(encoding)
    m.set_payload(body, charset=encoding)
    return m

def encode_header(str, encoding):
    """Attempt to encode a unicode header string word by word.  Let's try
    this out::

    ASCII strings should be unchanged::
        >>> encode_header(u'A simple subject', 'utf8')
        'A simple subject'

    Non-ASCII should be encoded::
        >>> encode_header(
        ... u'\\xc3\\xa9\\xc3\\xb8\\xe2\\x88\\x91\\xc3\\x9f\\xcf\\x80\\xc3\\xa5\\xe2\\x80\\xa0',
        ... 'utf8')
        '=?utf8?b?w4PCqcODwrjDosKIwpHDg8Kfw4/CgMODwqXDosKAwqA=?='

    A mix of the two should be nicely mixed::
        >>> encode_header(u'Je les d\\xc3\\xa9t\\xc3\\xa8ste, non?', 'utf8')
        'Je les =?utf8?b?ZMODwql0w4PCqHN0ZSw=?= non?'

    """
    cur_type = None
    last_type = None
    header_val = Header.Header()
    for word in str.split(' '):
        last_type = cur_type
        try:
            enc_word = word.encode('ascii')
            cur_type = 'ascii'
        except UnicodeEncodeError:
            enc_word = word.encode(encoding)
            cur_type = encoding
        header_val.append(enc_word, cur_type)

    return header_val.encode()

def construct_multipart(from_addr, to_addr, subject, body, html_body,
                        other_headers=None, encoding='utf8'):
    """All inputs to this method are expected to be unicode or ascii.

    We should be able to pass in some arbitrary unicode stuff and get back
    a sensible encoded message:

        >>> m = construct_multipart(u'test@example.com',
        ...     u'test@example.com',
        ...     u'Un Subj\xc3\xa9t',
        ...     u'A simple body with some non ascii t\xc3\xa9xt',
        ...     u'<p>A simple body with some non ascii t\xc3\xa9xt</p>',
        ...     other_headers = {'X-Test': u't\xc3\xa9st'})
        >>> print m.as_string() #doctest: +ELLIPSIS
        Content-Type: multipart/alternative; boundary="..."
        MIME-Version: 1.0
        From: test@example.com
        To: test@example.com
        Subject: Un =?utf8?b?U3ViasODwql0?=
        X-Test: =?utf8?b?dMODwqlzdA==?=
        <BLANKLINE>
        --...
        Content-Type: text/plain; charset="utf8"
        MIME-Version: 1.0
        Content-Transfer-Encoding: quoted-printable
        Content-Disposition: inline
        <BLANKLINE>
        A simple body with some non ascii t=C3=83=C2=A9xt
        --...
        Content-Type: text/html; charset="utf8"
        MIME-Version: 1.0
        Content-Transfer-Encoding: quoted-printable
        Content-Disposition: inline
        <BLANKLINE>
        <p>A simple body with some non ascii t=C3=83=C2=A9xt</p>
        --...
    """
    if other_headers is None:
        other_headers = {}
    m = MIMEMultipart('alternative')
    if 'From' not in other_headers:
        m['From'] = encode_header(from_addr, encoding)
    m['To'] = encode_header(to_addr, encoding)
    m['Subject'] = encode_header(subject, encoding)
    for key, val in other_headers.items():
        m[key] = encode_header(val, encoding)
    body = body.encode(encoding)
    txt = MIMEText(body,  _charset=encoding)
    txt['Content-Disposition'] = 'inline'
    m.attach(txt)
    html_body = html_body.encode(encoding)
    html = MIMEText(html_body, _subtype='html', _charset=encoding)
    html['Content-Disposition'] = 'inline'
    m.attach(html)
    return m



def construct_multipart_from_stx(from_addr, to_addr, subject, body,
                        other_headers=None, encoding='utf8'):
    """All inputs to this method are expected to be unicode or ascii.

    This converts the input text to html using the stx converter

        >>> m = construct_multipart_from_stx(u'test@example.com',
        ...     u'test@example.com',
        ...     u'Un Subj\xc3\xa9t',
        ...     u'A simple body with some non ascii t\xc3\xa9xt with "a link":http://www.example.com')
        >>> print m.as_string() #doctest: +ELLIPSIS
        Content-Type: multipart/alternative; boundary="..."
        MIME-Version: 1.0
        From: test@example.com
        To: test@example.com
        Subject: Un =?utf8?b?U3ViasODwql0?=
        <BLANKLINE>
        --...
        Content-Type: text/plain; charset="utf8"
        MIME-Version: 1.0
        Content-Transfer-Encoding: quoted-printable
        Content-Disposition: inline
        <BLANKLINE>
        A simple body with some non ascii t=C3=83=C2=A9xt with "a link":http://www.=
        example.com
        --...
        Content-Type: text/html; charset="utf8"
        MIME-Version: 1.0
        Content-Transfer-Encoding: quoted-printable
        Content-Disposition: inline
        <BLANKLINE>
        <p>A simple body with some non ascii t=C3=83=C2=A9xt with <a href=3D"http:/=
        /www.example.com">a link</a></p>
        <BLANKLINE>
        --...
    """
    if other_headers is None:
        other_headers = {}
    html = HTML(body, level=2, header=0)
    return construct_multipart(from_addr, to_addr, subject, body,
                               html, other_headers, encoding)



from zope.interface import implements
from zope.component import adapts
from plone.relations import interfaces as pr_interfaces
from plone.app.relations import interfaces
from zope.app.annotation.interfaces import IAnnotations
from persistent.mapping import PersistentMapping
from zc.relationship.interfaces import IBidirectionalRelationshipIndex

ANNOTATIONS_KEY = 'plone.app.relationship.relation_annotations'

class StateAnnotationsAdapter(object):
    implements(pr_interfaces.IStatefulRelationship)
    adapts(interfaces.IAnnotationsState)
    def __init__(self, rel):
        self.rel = rel
        self.annotations = IAnnotations(rel).setdefault(ANNOTATIONS_KEY,
                                                        PersistentMapping())
    @apply
    def state():
        def get(self):
            return self.annotations.get('state', None)
        def set(self, value):
            self.annotations['state'] = value
            if IBidirectionalRelationshipIndex.providedBy(self.rel.__parent__):
                self.rel.__parent__.reindex(self.rel)
        return property(get, set)

class ContextAnnotationsAdapter(object):
    implements(pr_interfaces.IContextAwareRelationship)
    adapts(interfaces.IAnnotationsContext)
    def __init__(self, rel):
        self.rel = rel
        self.annotations = IAnnotations(rel).setdefault(ANNOTATIONS_KEY,
                                                        PersistentMapping())
    def getContext(self):
        return self.annotations.get('context', None)

    def setContext(self, value):
        self.annotations['context'] = value
        if IBidirectionalRelationshipIndex.providedBy(self.rel.__parent__):
            self.rel.__parent__.reindex(self.rel)



from zope import interface
from zope.app.container.interfaces import IReadContainer
from zc.relationship.interfaces import IRelationship
from zc.relationship.interfaces import IRelationshipContainer
import zope.index.interfaces

_marker = []

class IComplexRelationship(IRelationship):
    """An asymmetric relationship using sources, a relationship
    type (predicate), and targets:
    e.g. (Alice and Bob) are (enemies) of (Eve and Mallory)"""

    relation = interface.Attribute(
        """uniode: the single relationship type of this relationship;
        usually contains the verb of the sentence.""")


class IContextAwareRelationship(IRelationship):
    """A specialized relationship which provides a limited context in which the
    relationship is defined (e.g. two people may have a specific relationship in
    the context of a particular project or department, but a different one in
    another)"""

    def getContext():
        """return a context for the relationship"""


class IStatefulRelationship(IRelationship):
    """A specialized relationship where the relation itself has a state that
    evolves over time.  In particular, one may want relationships where the
    subject initiates the relationship, but it must be approved by the object
    of the relationship.  Or a relationship like a friendship, which may
    evolve over time through different stages
    ('acquaintance', 'chum', 'friend', 'BFF!')"""

    state = interface.Attribute(
        """unicode: the current state of the relationship""")


class IComplexRelationshipContainer(IRelationshipContainer):
    """An IRelationshipContainer like the one from zc.relationship, but with
    a little more complex relationship modelling.  It also supports pluggable
    transitivity models in case the default relationship independent model
    is too simple."""

    def findTargets(source, relation=_marker, state=_marker, context=_marker,
                    maxDepth=1, minDepth=None, filter=None, transitivity=None):
        """See IBidirectionalRelationshipIndex: includes support for efficient
        filtering on state, context, and relationship.
        """

    def findSources(target, relation=_marker, state=_marker, context=_marker,
                    maxDepth=1, minDepth=None, filter=None, transitivity=None):
        """See IBidirectionalRelationshipIndex: includes support for efficient
        filtering on state, context, and relationship.
        """

    def isLinked(source=None, target=None, relation=_marker,
                 state=_marker, context=_marker, minDepth=None,
                 maxDepth=1, filter=None, transitivity=None):
        """See IBidirectionalRelationshipIndex: includes support for
        efficient filtering on state, context, and relationship."""

    def findRelationships(source=None, target=None, relation=_marker,
                          state=_marker, context=_marker, maxDepth=1,
                          minDepth=None, filter=None, transitivity=None):
        """See IBidirectionalRelationshipIndex: includes support for efficient
        filtering on state, context, and relationship.
        """

    def findTargetTokens(source, relation=_marker, state=_marker,
                         context=_marker, maxDepth=1, minDepth=None,
                         filter=None, transitivity=None):
        """See IBidirectionalRelationshipIndex: includes support for efficient
        filtering on state, context, and relationship.
        """

    def findSourceTokens(source, relation=_marker, state=_marker,
                         context=_marker, maxDepth=1, minDepth=None,
                         filter=None, transitivity=None):
        """See IBidirectionalRelationshipIndex: includes support for efficient
        filtering on state, context, and relationship.
        """

    def findRelationshipTokens(source, relation=_marker, state=_marker,
                               context=_marker, maxDepth=1, minDepth=None,
                               filter=None, transitivity=None):
        """See IBidirectionalRelationshipIndex: includes support for efficient
        filtering on state, context, and relationship.
        """



from itertools import izip, repeat
from Globals import InitializeClass
from Acquisition import aq_inner, aq_parent
from AccessControl import ClassSecurityInfo
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from zope.component import getAdapters, adapts
from zope.interface import implements

from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PlonePAS.interfaces.plugins import ILocalRolesPlugin

from plone.app.relations.interfaces import ILocalRoleProvider
from plone.app.relations.interfaces import IFactoryTempFolder

manage_addRelationshipLocalRoleManagerForm = PageTemplateFile(
        "AddLocalRoleManagerForm.pt", globals(),
        __name__="manage_addRelationshipRoleManagerForm")

def manage_addRelationshipLocalRoleManager(dispatcher, id, title=None,
                                           REQUEST=None):
    """Add a ProjectLocalRoleManager to a Pluggable Authentication Services."""
    lrm = RelationshipLocalRoleManager(id, title)
    dispatcher._setObject(lrm.getId(), lrm)

    if REQUEST is not None:
        REQUEST.RESPONSE.redirect(
                '%s/manage_workspace?manage_tabs_message='
                'RelationshipLocalRoleManager+added.'%dispatcher.absolute_url())

class RelationshipLocalRoleManager(BasePlugin):
    """Though it's aimed for use in defining local roles based on relationships,
    this local role manager is completely generic, and simply uses the Zope 3
    CA to look up local role providing components via adaptation.  Essentially
    this is the result of my being to lazy to write individual persistent
    plugins for every different usecase which may come up.


    First we need to make and register an adapter to provide some roles::

        >>> from zope.interface import implements, Interface
        >>> from zope.component import adapts
        >>> class SimpleLocalRoleProvider(object):
        ...     adapts(Interface)
        ...     implements(ILocalRoleProvider)
        ...
        ...     def __init__(self, context):
        ...         self.context = context
        ...
        ...     def getRoles(self, user):
        ...         '''Grant everyone the 'Foo' role'''
        ...         return ('Foo',)
        ...
        ...     def getAllRoles(self):
        ...         '''In the real world we would enumerate all users and
        ...         grant the 'Foo' role to each, but we won't'''
        ...         yield ('bogus_user', ('Foo',))

        >>> from zope.component import provideAdapter
        >>> provideAdapter(SimpleLocalRoleProvider)


    We need an object to adapt, we require nothing of this object,
    except it must be adaptable (e.g. have an interface)::

        >>> class DummyObject(object):
        ...     implements(Interface)
        >>> ob = DummyObject()

    And we need some users that we'll check the permissions of::

        >>> class DummyUser(object):
        ...     def __init__(self, uid, group_ids=()):
        ...         self.id = uid
        ...         self._groups = group_ids
        ...
        ...     def getId(self):
        ...         return self.id
        ...
        ...     def _check_context(self, obj):
        ...         return True
        ...
        ...     def getGroups(self):
        ...         return self._groups
        ...
        >>> user1 = DummyUser('bogus_user')
        >>> user2 = DummyUser('bogus_user2')

    Now we're ready to make one of our RoleManagers and try it out.
    First we'll verify that our users have the 'Foo' role, then we'll
    make sure they can access objects which require that role, but not
    others::

        >>> rm = RelationshipLocalRoleManager('rm', 'A Role Manager')
        >>> rm.getRolesInContext(user1, ob)
        ['Foo']
        >>> rm.checkLocalRolesAllowed(user1, ob, ['Bar', 'Foo', 'Baz'])
        1
        >>> rm.checkLocalRolesAllowed(user1, ob, ['Bar', 'Baz']) is None
        True
        >>> rm.getAllLocalRolesInContext(ob)
        {'bogus_user': set(['Foo'])}


    Multiple Role Providers
    -----------------------

    It is a bit more interesting when we have more than one adapter
    registered.  We register it with a name so that it supplements,
    rather than conflict with or override the existing adapter::

        >>> class LessSimpleLocalRoleProvider(SimpleLocalRoleProvider):
        ...     userid = 'bogus_user2'
        ...     roles = ('Foo', 'Baz')
        ...     def getRoles(self, userid):
        ...         '''Grant bogus_user2 the 'Foo' and 'Baz' roles'''
        ...         if userid == self.userid:
        ...             return self.roles
        ...         return ()
        ...
        ...     def getAllRoles(self):
        ...         yield (self.userid, self.roles)

        >>> provideAdapter(LessSimpleLocalRoleProvider, name='adapter2')

   This should have no effect on our first user::

        >>> rm.getRolesInContext(user1, ob)
        ['Foo']
        >>> rm.checkLocalRolesAllowed(user1, ob, ['Bar', 'Foo', 'Baz'])
        1
        >>> rm.checkLocalRolesAllowed(user1, ob, ['Bar', 'Baz']) is None
        True
        >>> rm.getAllLocalRolesInContext(ob)
        {'bogus_user2': set(['Foo', 'Baz']), 'bogus_user': set(['Foo'])}

    But our second user notices the change, note that even though two
    of our local role providers grant the role 'Foo', it is not duplicated::

        >>> rm.getRolesInContext(user2, ob)
        ['Foo', 'Baz']
        >>> rm.checkLocalRolesAllowed(user2, ob, ['Bar', 'Foo', 'Baz'])
        1
        >>> rm.checkLocalRolesAllowed(user2, ob, ['Bar', 'Baz'])
        1
        >>> rm.checkLocalRolesAllowed(user2, ob, ['Bar']) is None
        True


    Role Acquisition and Blocking
    -----------------------------

    This plugin will acquire role definitions from parent objects,
    unless explicitly blocked.  To test this, we need some objects
    which support acquisition::

        >>> from Acquisition import Implicit
        >>> class DummyImplicit(DummyObject, Implicit):
        ...     def stupid_method(self):
        ...         return 1
        >>> root = DummyImplicit()
        >>> next = DummyImplicit().__of__(root)
        >>> last = DummyImplicit().__of__(next)
        >>> other = DummyImplicit().__of__(root)

    So we now have /root/next/last and /root/other, we'll create and
    register special adapters for our next and other objects.

        >>> class ISpecial1(Interface):
        ...     pass
        >>> class ISpecial2(Interface):
        ...     pass
        >>> from zope.interface import directlyProvides
        >>> directlyProvides(next, ISpecial1)
        >>> directlyProvides(other, ISpecial2)
        >>> class Adapter1(LessSimpleLocalRoleProvider):
        ...     adapts(ISpecial1)
        ...     userid = 'bogus_user'
        ...     roles = ('Bar',)
        >>> class Adapter2(LessSimpleLocalRoleProvider):
        ...     adapts(ISpecial2)
        ...     userid = 'bogus_user3'
        ...     roles = ('Foobar',)
        >>> user3 = DummyUser('bogus_user3')

    We'll register these to override the existing unnamed adapter:

        >>> provideAdapter(Adapter1)
        >>> provideAdapter(Adapter2)

    Now we can show how acquisition of roles works, first we look at the
    'last' item, which should have roles provided by
    SimpleLocalRoleProvider, and LessSimpleLocalRoleProvider, as well
    as acquired from Adapter1 on 'next':

        >>> rm.getRolesInContext(user1, last)
        ['Foo', 'Bar']

        >>> rm.getRolesInContext(user2, last)
        ['Foo', 'Baz']

    If we look at the parent, we get the same results, because the
    SimpleLocalRoleProvider adapter also applies to the 'root'
    object. However, if we enable local role blocking on 'next' we
    won't see the roles from the 'root'::

        >>> rm.getRolesInContext(user1, next)
        ['Foo', 'Bar']
        >>> next.__ac_local_roles_block__ = True
        >>> rm.getRolesInContext(user1, next)
        ['Bar']

    The checkLocalRolesAllowed and getAllLocalRolesInContext methods
    take acquisition and blocking into account as well::

        >>> rm.checkLocalRolesAllowed(user1, last,  ['Bar'])
        1
        >>> rm.checkLocalRolesAllowed(user1, next,  ['Foo', 'Baz']) is None
        True
        >>> rm.getAllLocalRolesInContext(last)
        {'bogus_user2': set(['Foo', 'Baz']), 'bogus_user': set(['Foo', 'Bar'])}

    It's important to note, that roles are acquired only by
    containment.  Additional wrapping cannot change the security on an
    object.  For example if we were to wrap 'last' in the context of
    other, which provides a special role for 'user3', we should see no
    effect::

        >>> rm.getRolesInContext(user3, last)
        ['Foo']
        >>> rm.getRolesInContext(user3, other)
        ['Foobar', 'Foo']
        >>> rm.getRolesInContext(user3, last.__of__(other))
        ['Foo']

    It's also important that methods of objects yield the same local
    roles as the objects would

        >>> rm.getRolesInContext(user3, other.stupid_method)
        ['Foobar', 'Foo']

    Group Support
    -------------

    This plugin also handles roles granted to user groups, calling up
    the adapters to get roles for any groups the user might belong
    to::

        >>> user4 = DummyUser('bogus_user4', ('Group1', 'Group2'))
        >>> user4.getGroups()
        ('Group1', 'Group2')
        >>> rm.getRolesInContext(user4, last)
        ['Foo']
        >>> class Adapter3(LessSimpleLocalRoleProvider):
        ...     userid = 'Group2'
        ...     roles = ('Foobar',)

        >>> provideAdapter(Adapter3, name='group_adapter')
        >>> rm.getRolesInContext(user4, last)
        ['Foobar', 'Foo']


    Wrong User Folder
    -----------------

    Finally, to ensure full test coverage, we provide a user object
    which pretends to be wrapped in such a way that the user folder
    does not recognize it.  We check that it always gets an empty set
    of roles and a special 0 value when checking access::

        >>> class BadUser(DummyUser):
        ...     def _check_context(self, obj):
        ...         return False
        >>> bad_user = BadUser('bad_user')
        >>> rm.getRolesInContext(bad_user, ob)
        []
        >>> rm.checkLocalRolesAllowed(bad_user, ob, ['Bar', 'Foo', 'Baz'])
        0

    """
    meta_type = "Relationship Roles Manager"
    security  = ClassSecurityInfo()

    def __init__(self, id, title=''):
        self.id = id
        self.title = title

    #
    # ILocalRolesPlugin implementation
    #

    def _getAdapters(self, obj):
        adapters = getAdapters((obj,), ILocalRoleProvider)
        # this is sequence of tuples of the form (name, adapter),
        # we don't really care about the names
        return (a[1] for a in adapters)

    def _parent_chain(self, obj):
        """Iterate over the containment chain, stopping if we hit a
        local role blocker"""
        while obj is not None:
            obj = aq_inner(obj)
            yield obj
            if getattr(obj, '__ac_local_roles_block__', None):
                raise StopIteration
            new = aq_parent(obj)
            # if the obj is a method we get the class
            obj = getattr(obj, 'im_self', new)

    def _get_principal_ids(self, user):
        """Returns a list of the ids of all involved security
        principals: the user and all groups that they belong
        to. (Note: recursive groups are not yet supported"""
        principal_ids = list(user.getGroups())
        principal_ids.insert(0, user.getId())
        return principal_ids

    security.declarePrivate("getRolesInContext")
    def getRolesInContext(self, user, object):
        # we combine the permission of the user with those of the
        # gorups she belongs to
        principal_ids = self._get_principal_ids(user)
        roles = set()
        if user._check_context(object):
            for obj in self._parent_chain(object):
                for a in self._getAdapters(obj):
                    for pid in principal_ids:
                        roles.update(a.getRoles(pid))
        return list(roles)

    security.declarePrivate("checkLocalRolesAllowed")
    def checkLocalRolesAllowed(self, user, object, object_roles):
        """Checks if the user has one of the specified roles in the
        given context, short circuits when the first provider granting
        one of the roles is found."""
        check_roles = dict(izip(object_roles, repeat(True)))
        principal_ids = self._get_principal_ids(user)
        if not user._check_context(object):
            return 0
        for obj in self._parent_chain(object):
            for a in self._getAdapters(obj):
                for pid in principal_ids:
                    roles = a.getRoles(pid)
                    for role in check_roles:
                        if role in roles:
                            return 1
        return None

    security.declarePrivate("getAllLocalRolesInContext")
    def getAllLocalRolesInContext(self, object):
        rolemap = {}
        for obj in self._parent_chain(object):
            for a in self._getAdapters(obj):
                iter_roles = a.getAllRoles()
                for principal, roles in iter_roles:
                    rolemap.setdefault(principal, set()).update(roles)

        return rolemap

classImplements(RelationshipLocalRoleManager, ILocalRolesPlugin)
InitializeClass(RelationshipLocalRoleManager)

from Acquisition import aq_inner, aq_parent
from Products.CMFCore.utils import getToolByName

class FactoryTempFolderProvider(object):
    """A simple local role provider which just gathers the roles from
    the desired context"""
    adapts(IFactoryTempFolder)
    implements(ILocalRoleProvider)

    def __init__(self, obj):
        self.folder = obj

    def getRoles(self, principal_id):
        uf = aq_inner(getToolByName(self.folder, 'acl_users'))
        user = aq_inner(uf.getUserById(principal_id, default=None))
        # use the folder we are creating in as role generating context
        source = aq_parent(aq_parent(self.folder))
        if user is not None:
            return user.getRolesInContext(source)
        else:
            return []

    def getAllRoles(self):
        # This should not be used in any meaningful way, so we'll make it cheap
        return {}



from Acquisition import Implicit, aq_parent, aq_inner, aq_base
import persistent
from BTrees.IIBTree import IITreeSet
from zope import interface
from zope.component import adapts
from zope.component import getUtility
from zope.app.intid.interfaces import IIntIds
from zc.relationship.shared import Relationship as RelationshipBase
from zc.relationship import interfaces as zc_interfaces
from plone.relations import interfaces

class IntIdSubObjectWrapper(IITreeSet):

    def __init__(self, initial):
        super(IntIdSubObjectWrapper, self).__init__(self._get_id(o) for o in
                                                                       initial)

    def _get_id(self, obj):
        resolver = getattr(self, '_v_intid_resolver', None)
        if resolver is None:
            resolver = self._v_intid_resolver = getUtility(IIntIds).getId
        return resolver(obj)

    def _get_object(self, intid):
        resolver = getattr(self, '_v_object_resolver', None)
        if resolver is None:
            resolver = self._v_object_resolver = getUtility(IIntIds).getObject
        return resolver(intid)

    def __iter__(self):
        for item in IITreeSet.__iter__(self):
            yield self._get_object(item)

    def __contains__(self, obj):
        return IITreeSet.__contains__(self, self._get_id(obj))

    def remove(self, obj):
        return IITreeSet.remove(self, self._get_id(obj))


class Relationship(RelationshipBase):
    interface.implements(interfaces.IRelationship)
    _relation = None

    def __init__(self, sources, targets, relation=None):
        self._sources = IntIdSubObjectWrapper(sources)
        self._targets = IntIdSubObjectWrapper(targets)
        # use the adapter to set the relation type
        if relation is not None:
            interfaces.IComplexRelationship(self).relation = relation
    @apply
    def sources():
        def get(self):
            return self._sources
        def set(self, value):
            self._sources = IntIdSubObjectWrapper(value)
            if zc_interfaces.IBidirectionalRelationshipIndex.providedBy(
                self.__parent__):
                self.__parent__.reindex(self)
        return property(get, set)

    @apply
    def targets():
        def get(self):
            return self._targets
        def set(self, value):
            self._targets = IntIdSubObjectWrapper(value)
            if zc_interfaces.IBidirectionalRelationshipIndex.providedBy(
                self.__parent__):
                self.__parent__.reindex(self)
        return property(get, set)

    def __repr__(self):
        try:
            sources = tuple(self.sources)
        except AttributeError:
            sources = '<Missing>'
        try:
            targets = tuple(self.targets)
        except AttributeError:
            targets = '<Missing>'
        return '<Relationship %r from %r to %r>' % (
                                 interfaces.IComplexRelationship(self).relation,
                                 sources, targets)

# A version of the class with acquisition in case things need to be
# acquirable from the relationship.
class Z2Relationship(Relationship, Implicit):
    pass

class ComplexRelationshipAdapter(object):
    interface.implements(interfaces.IComplexRelationship)
    adapts(interfaces.IRelationship)
    def __init__(self, rel):
        self.rel = rel

    @apply
    def relation():
        def get(self):
            return getattr(self.rel, '_relation', None)
        def set(self, value):
            self.rel._relation = value
            if zc_interfaces.IBidirectionalRelationshipIndex.providedBy(
                self.rel.__parent__):
                self.rel.__parent__.reindex(self.rel)
        return property(get, set)



import random
from Acquisition import Explicit, aq_parent, aq_inner, aq_base
from BTrees import OIBTree
import zope.app.container.btree
from zope.interface import implements
from zope.event import notify
from zope.app.container.contained import ObjectRemovedEvent
from zc.relationship import index
from zc.relationship.interfaces import ICircularRelationshipPath
from zc.relationship.shared import ResolvingFilter
from zc.relationship.shared import minDepthFilter
from zc.relationship.shared import Container, AbstractContainer
from plone.relations import interfaces

_marker = interfaces._marker

def _update_query(query, relation, state, context):
    if relation != _marker:
        query['relation'] = relation
    if state != _marker:
        query['state'] = state
    if context != _marker:
        query['context'] = context

# A dumper, which is also a loader, for simple strings
def str_dump(obj, index, cache):
    return obj
str_load = str_dump

def _set_state(obj, dump, load, family):
    if dump is not None:
        obj['dump'] = dump
    if load is not None:
        obj['load'] = load
    if family is not None:
        obj['btree'] = family

# RelationshipContainers whose relationships need to interact with the
# security machinery of Zope 2 should inherit from one of the
# Acquisition base classes.
class RelationshipContainer(Container):

    # Set aatus so that we can retrieve entries, which are never protected.
    # The core methods are all protected via zcml declarations.
    # This is only necessary because of the use of restrictedTraverse in
    # five.intid's keyreference lookup. (see OFS.Folder)
    __allow_access_to_unprotected_subobjects__ = 1
    __name__ = 'relations'

    implements(interfaces.IComplexRelationshipContainer)
    def __init__(self,
                 dumpSource=None, loadSource=None, sourceFamily=None,
                 dumpTarget=None, loadTarget=None, targetFamily=None,
                 dumpRelation=None, loadRelation=None, relationFamily=None,
                 dumpState=None, loadState=None, stateFamily=None,
                 dumpContext=None, loadContext=None, contextFamily=None,
                 **kwargs):
        # Override the AbstractContainer __init__
        source = {'element': interfaces.IRelationship['sources'],
                  'name': 'source', 'multiple': True}
        target = {'element': interfaces.IRelationship['targets'],
                  'name': 'target', 'multiple': True}
        relation = {'element': interfaces.IComplexRelationship['relation'],
                    'name': 'relation', 'dump': str_dump, 'load': str_load,
                    'btree': OIBTree}
        state = {'element': interfaces.IStatefulRelationship['state'],
                 'name': 'state', 'dump': str_dump, 'load': str_load,
                 'btree': OIBTree}
        context = {'element': interfaces.IContextAwareRelationship['getContext'],
                   'name': 'context'}

        # set all the non-string indices to use the source dump/load/storage if
        # specified
        if dumpSource is not None:
            target['dump'] = source['dump'] = context['dump'] = dumpSource
        if loadSource is not None:
            target['load'] = source['load'] = context['load'] = loadSource
        if sourceFamily is not None:
            target['btree'] = source['btree'] = context['btree'] = sourceFamily

        # set the more specific dump/load/family specs to their indices
        _set_state(target, dumpTarget, loadTarget, targetFamily)
        _set_state(relation, dumpRelation, loadRelation, relationFamily)
        _set_state(state, dumpState, loadState, stateFamily)
        _set_state(context, dumpContext, loadContext, contextFamily)

        ix = index.Index(
            (source, target, relation, state, context),
            index.TransposingTransitiveQueriesFactory('source', 'target'),
            **kwargs)
        self.relationIndex = ix
        ix.__parent__ = self
        zope.app.container.btree.BTreeContainer.__init__(self)

    def reindex(self, object):
        assert object.__parent__ is self
        self.relationIndex.index(object)

    def findTargets(self, source, relation=_marker, state=_marker,
                    context=_marker, maxDepth=1, minDepth=None, filter=None,
                    transitivity=None):
        query = {'source': source}
        _update_query(query, relation, state, context)
        return self.relationIndex.findValues(
            'target', self.relationIndex.tokenizeQuery(query),
            maxDepth, filter and ResolvingFilter(filter, self),
            targetFilter=minDepthFilter(minDepth),
            transitiveQueriesFactory=transitivity)

    def findSources(self, target, relation=_marker, state=_marker,
                    context=_marker, maxDepth=1, minDepth=None, filter=None,
                    transitivity=None):
        query = {'target': target}
        _update_query(query, relation, state, context)
        return self.relationIndex.findValues(
            'source', self.relationIndex.tokenizeQuery(query),
            maxDepth, filter and ResolvingFilter(filter, self),
            targetFilter=minDepthFilter(minDepth),
            transitiveQueriesFactory=transitivity)

    def findTargetTokens(self, source, relation=_marker, state=_marker,
                         context=_marker, maxDepth=1, minDepth=None,
                         filter=None, transitivity=None):
        query = {'source': source}
        _update_query(query, relation, state, context)
        return self.relationIndex.findValueTokens(
            'target', self.relationIndex.tokenizeQuery(query),
            maxDepth, filter and ResolvingFilter(filter, self),
            targetFilter=minDepthFilter(minDepth),
            transitiveQueriesFactory=transitivity)

    def findSourceTokens(self, target, relation=_marker, state=_marker,
                         context=_marker, maxDepth=1, minDepth=None,
                         filter=None, transitivity=None):
        query = {'target': target}
        _update_query(query, relation, state, context)
        return self.relationIndex.findValueTokens(
            'source', self.relationIndex.tokenizeQuery(query),
            maxDepth, filter and ResolvingFilter(filter, self),
            targetFilter=minDepthFilter(minDepth),
            transitiveQueriesFactory=transitivity)

    def isLinked(self, source=None, target=None, relation=_marker,
                 state=_marker, context=_marker, maxDepth=1, minDepth=None,
                 filter=None, transitivity=None):
        tokenize = self.relationIndex.tokenizeQuery
        query = {}
        _update_query(query, relation, state, context)
        if source is not None:
            query['source'] = source
            return self.relationIndex.isLinked(
                tokenize(query),
                maxDepth, filter and ResolvingFilter(filter, self),
                target and tokenize({'target': target}),
                targetFilter=minDepthFilter(minDepth),
                transitiveQueriesFactory=transitivity)
        elif target is not None:
            query['target'] = target
            return self.relationIndex.isLinked(
                tokenize(query),
                maxDepth, filter and ResolvingFilter(filter, self),
                targetFilter=minDepthFilter(minDepth),
                transitiveQueriesFactory=transitivity)
        else:
            raise ValueError(
                'at least one of `source` and `target` must be provided')

    def findRelationshipTokens(self, source=None, target=None, relation=_marker,
                               state=_marker, context=_marker, maxDepth=1,
                               minDepth=None, filter=None, transitivity=None):
        tokenize = self.relationIndex.tokenizeQuery
        query = {}
        _update_query(query, relation, state, context)
        if source is not None:
            query['source'] = source
            res = self.relationIndex.findRelationshipTokenChains(
                tokenize(query),
                maxDepth, filter and ResolvingFilter(filter, self),
                target and tokenize({'target': target}),
                targetFilter=minDepthFilter(minDepth),
                transitiveQueriesFactory=transitivity)
            return self._forward(res)
        elif target is not None:
            query['target'] = target
            res = self.relationIndex.findRelationshipTokenChains(
                tokenize(query),
                maxDepth, filter and ResolvingFilter(filter, self),
                targetFilter=minDepthFilter(minDepth),
                transitiveQueriesFactory=transitivity)
            return self._reverse(res)
        elif relation is not _marker or context is not _marker:
            res = self.relationIndex.findRelationshipTokenChains(
                tokenize(query),
                maxDepth, filter and ResolvingFilter(filter, self),
                targetFilter=minDepthFilter(minDepth),
                transitiveQueriesFactory=transitivity)
            return self._reverse(res)
        else:
            raise ValueError(
                'at least one of `source`, `target`, `relation`, or `context` '
                'must be provided')

    def findRelationships(self, source=None, target=None, relation=_marker,
                          state=_marker, context=_marker, maxDepth=1,
                          minDepth=None, filter=None, transitivity=None):
        return self._resolveRelationshipChains(
            self.findRelationshipTokens(
                source, target, relation, state, context, maxDepth,
                minDepth, filter, transitivity=transitivity))

    # We cannot use _ in names because it causes random-ish problems in Zope2
    def _generate_id(self, relationship):
        return ''.join(random.sample(
            "abcdefghijklmnopqrtstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-1234567890",
            30)) # somewhat less than 64 ** 30 variations (64*63*...*35)

# A version of the class with acquisition added so that Zope 2 security
# checks may be performed on the contained objects.  Note: when this container
# is acquisition wrapped, you will not be able to iterate over its keys using
# list(container) or the 'in' operator due to some odd acquisition behavior.
class Z2RelationshipContainer(RelationshipContainer, Explicit):

    def __init__(self, name=None, *args, **kw):
            self.__name__ = name
            super(Z2RelationshipContainer, self).__init__(*args, **kw)

    def manage_fixupOwnershipAfterAdd(self):
        pass

    def __getitem__(self, key):
        # acquisition wrap the output
        val = super(Z2RelationshipContainer, self).__getitem__(key)
        if hasattr(val, '__of__'):
            val = val.__of__(self)
        return val

    def get(self, key, default=None):
        # acquisition wrap the output
        val = super(Z2RelationshipContainer, self).get(key, default)
        if hasattr(val, '__of__'):
            val = val.__of__(self)
        return val

    def reindex(self, object):
        assert aq_base(object.__parent__) is aq_base(self)
        self.relationIndex.index(object)

    def remove(self, object):
        key = object.__name__
        if aq_base(self[key]) is not aq_base(object):
            raise ValueError("Relationship is not stored as its __name__")
        self.relationIndex.unindex(object)
        # we need to manually send the ObjectRemovedEvent, because acquisition
        # wrapping prevents it from firing automatically.
        notify(ObjectRemovedEvent(object, self, key))
        super(AbstractContainer, self).__delitem__(key)



from plone.relations.interfaces import IComplexRelationshipContainer, _marker
from plone.relations.relationships import Z2Relationship
from plone.relations.lazylist import lazyresolver
from plone.app.relations import interfaces as pa_interfaces
from zope.component import adapts, getUtility
from zope.interface import implements, alsoProvides
from persistent import IPersistent


class RelationshipSource(object):
    """A basic implementation of IRelationshipSource based on the container
    from plone.relations, this package registers it as a named utility
    called ``relations``"""
    implements(pa_interfaces.IRelationshipSource)
    adapts(IPersistent)
    _name = 'relations'

    def __init__(self, source):
        self.source = source
        # always use the context of the source object for utility lookup
        self.util = getUtility(IComplexRelationshipContainer, name=self._name,
                               context=source)
        self._resolver = self.util.relationIndex.resolveValueTokens

    def _source_resolver(self, value):
        return self._resolver((value,), 'source').next()

    def createRelationship(self, targets, relation=None, interfaces=(),
                           rel_factory=None, default_deletion=True):
        """See interface"""
        if rel_factory is None:
            rel_factory = Z2Relationship
        if not isinstance(targets, (list, tuple)):
            targets = (targets,)
        rel = rel_factory((self.source,), targets, relation=relation)
        alsoProvides(rel, *interfaces)
        if default_deletion:
            alsoProvides(rel, pa_interfaces.IDefaultDeletion)
        self.util.add(rel)
        # retrieve the object from the container
        return self.util[rel.__name__]

    def getRelationships(self, target=None, relation=_marker, state=_marker,
                         context=_marker, rel_filter=None):
        """See interface"""
        rels = self.util.findRelationships(self.source, target, relation,
                                           state, context, filter=rel_filter)
        for chain in rels:
            yield chain[0]

    def deleteRelationship(self, target=None, relation=_marker, state=_marker,
                           context=_marker, rel_filter=None, multiple=False,
                           remove_all_targets=False, ignore_missing=False):
        """See interface"""
        if target is None:
            # No target specified implies that all targets should be affected
            remove_all_targets = True
        rels = self.getRelationships(target, relation, state, context,
                                     rel_filter)
        # Resolve the relationships to check their length.
        rels = list(rels)
        if not rels and not ignore_missing:
            raise pa_interfaces.NoResultsError
        if len(rels) > 1 and not multiple:
            raise pa_interfaces.TooManyResultsError
        for rel in rels:
            if target is not None:
                # Sanity check
                assert(target in rel.targets)
            # If there are multiple sources delete ourself from the list of
            # sources if remove_all_targets is set. Otherwise, raise an error.
            if len(rel.sources) > 1:
                if remove_all_targets or len(rel.targets) == 1:
                    rel.sources.remove(self.source)
                    self.util.reindex(rel)
                else:
                    raise pa_interfaces.TooManyResultsError, "One of the "\
                          "relationships to be deleted has multiple sources "\
                          "and targets."
            # If there are multiple targets and remove_all_targets is
            # set, remove them all.  Otherwise, remove only the
            # specified target.
            elif len(rel.targets) > 1 and not remove_all_targets:
                rel.targets.remove(target)
                self.util.reindex(rel)
            else:
                self.util.remove(rel)


    def isLinked(self, target=None, relation=_marker, state=_marker,
                 context=_marker, rel_filter=None, maxDepth=1,
                 minDepth=None, transitivity=None):
        """See interface"""
        return self.util.isLinked(self.source, target, relation, state,
                                  context, maxDepth=maxDepth,
                                  minDepth=minDepth,
                                  filter=rel_filter, transitivity=transitivity)

    def getRelationshipChains(self, target=None, relation=_marker,
                               state=_marker, context=_marker, rel_filter=None,
                               maxDepth=1, minDepth=None, transitivity=None):
        """See interface"""
        return self.util.findRelationships(self.source, target, relation,
                                           state, context, maxDepth=maxDepth,
                                           minDepth=minDepth,
                                           filter=rel_filter,
                                           transitivity=transitivity)

    @lazyresolver(resolver_name='_source_resolver')
    def getTargets(self, relation=_marker, state=_marker, context=_marker,
                    rel_filter=None, maxDepth=1, minDepth=None,
                    transitivity=None):
        """See interface"""
        return self.util.findTargetTokens(self.source, relation, state,
                                          context, maxDepth=maxDepth,
                                          minDepth=minDepth,
                                          filter=rel_filter,
                                          transitivity=transitivity)




from setuptools import setup, find_packages

version = '1.0b1'

setup(name='plone.relations',
      version=version,
      description="Tools for defining and querying complex relationships between objects",
      long_description=open("plone/relations/README.txt").read(),
      classifiers=[
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='references relationships plone',
      author='Alec Mitchell',
      author_email='apm13@columbia.edu',
      url='http://svn.plone.org/svn/plone/plone.relations',
      license='GPL with container.txt covered by ZPL and owned by Zope Corp.',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['plone'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          "zc.relationship",
          "five.intid",
      ],
      )



from itertools import chain
from zope.component import getUtility
from zope.app.intid.interfaces import IIntIds

class LazyList(object):
    """A class that makes an iterator look like a sequence,
    it only unspools the iterator as needed (no infinite iterators please),
    to perform its operations.

    Let's test this out, first we create lazy lists from existing
    iterators::

        >>> iter1 = xrange(20)
        >>> list1 = LazyList(iter1)
        >>> def iter2_func():
        ...     for i in xrange(20):
        ...         yield i
        >>> list2 = LazyList(iter2_func())

    First we ensure that we have a working len(), and that calling it
    doesn't result in unspooling the iterator unless necessary::

        >>> len(list1)
        20
        >>> list1._list is None
        True
        >>> list1[5]
        5
        >>> list1._list
        [0, 1, 2, 3, 4, 5]
        >>> list1[2:4]
        [2, 3]
        >>> list1._list
        [0, 1, 2, 3, 4, 5]
        >>> list1[-3]
        17
        >>> len(list1._list)
        20

    Now the iterator is entirely unspooled::

        >>> [i for i in list1._iter]
        []
        >>> list1._list == list(list1)
        True

    Our generator function behaves the same, except that using len() unspools
    it entirely::

        >>> list2[:5]
        [0, 1, 2, 3, 4]
        >>> list2._list
        [0, 1, 2, 3, 4]
        >>> list2[4:7]
        [4, 5, 6]
        >>> list2._list
        [0, 1, 2, 3, 4, 5, 6]
        >>> len(list2)
        20
        >>> len(list2._list)
        20

    We can also add these lists, if they are both fully unspooled we
    get a normal list::

        >>> list3 = list2 + list1
        >>> isinstance(list3, list)
        True
        >>> len(list3)
        40

    Otherwise we get another LazyList, note that when we slice past
    the end of the list the whole list gets resolved as expected::

        >>> list4 = LazyList(iter2_func())
        >>> list5 = list4 + list1
        >>> isinstance(list5, LazyList)
        True
        >>> list5[19:21]
        [19, 0]
        >>> len(list5._list)
        21
        >>> list5[:53] == list(list5)
        True
        >>> len(list5._list)
        40

    We can also "reverse" the list in place, which will unspool the
    whole thing::

        >>> list6 = LazyList(iter2_func())
        >>> list6[6]
        6
        >>> list6.reverse()
        >>> list6[6]
        13
        >>> len(list6._list)
        20

    We can also apply the reversed built-in to the list::

        >>> list7 = LazyList(iter2_func())
        >>> r = reversed(list7)
        >>> r.next()
        19
        >>> list7[0]
        0
        >>> len(list7._list)
        20

    We also have truth testing which doesn't unspool the list::

        >>> list8 = LazyList(iter2_func())
        >>> bool(list8)
        True
        >>> len(list8._list)
        1

   We can also handle regular lists as input::

        >>> list9 = LazyList([1,2,3,4,5])
        >>> len(list9._list)
        5
        >>> list9[0]
        1
        >>> list9[1:2]
        [2]

    """
    __allow_access_to_unprotected_subobjects__ = True
    _finished = False
    _list = None

    def __init__(self, initlist):
        self._orig_iter = initlist
        self._iter = iter(initlist)
        if isinstance(initlist, (list, tuple)):
            # skip extra processing when we get a list or tuple
            self._list = list(initlist)
            self._finished = True

    @property
    def _pos(self):
        try:
            return len(self._list) - 1
        except TypeError:
            # the list isn't initialized
            return -1

    def __iter__(self):
        # iterate over our internal list first
        # then the iterator
        if self._list is None:
            self._list = []
        for item in self._list:
            yield item
        if not self._finished:
            for item in self._iter:
                self._list.append(item)
                yield item
            else:
                # We completed the final iteration
                self._finished = True

    def _extract_to(self, end):
        if self._finished or (self._pos >= end and end >= 0):
            pass
        elif end < 0:
            # We need to unwrap the whole thing
            thelist = [i for i in self] # invokes __iter__ which sets _list
        else:
            num_iters = end - self._pos
            # continue where we left off
            if self._list is None:
                self._list = []
            for i, item in enumerate(self._iter):
                self._list.append(item)
                if i+1 >= num_iters:
                    break
            else:
                self._finished = True
        return self._list

    def __len__(self):
        # use the native __len__ if available
        if hasattr(self._orig_iter, '__len__'):
            return len(self._orig_iter)
        # Otherwise unspool the whole thing
        return len(self._extract_to(-1))

    def __getitem__(self, i):
        return self._extract_to(i)[i]

    def __getslice__(self, i, j):
        return self._extract_to(j-1)[i:j]

    def __reversed__(self):
        return reversed(self._extract_to(-1))

    def __add__(self, other):
        if self._finished and getattr(other, '_finished', False):
            return self._list + other._list
        return self.__class__(chain(self.__iter__(), other.__iter__()))

    def reverse(self):
        self._extract_to(-1).reverse()

    def __nonzero__(self):
        """Override this to avoid unspooling the whole list for a len check"""
        return bool(self._extract_to(0))


class LazyResolvingList(object):
    """A sequence class that acts like a list by lazily resolving
    a set of tokens, assumes tokens are intids, a custom resolver method
    can be passed in, it must take accept a single token and return
    a value.

    To see it in action we make a trivial resolver with a dictionary lookup::

        >>> TABLE = {0: 'a', 1: 'b', 2: 'c', 3: 'd', 4: 'e', 5: 'f',
        ...          6: 'g', 7: 'h', 8: 'i', 9: 'j'}
        >>> resolver = TABLE.get
        >>> counter = 0
        >>> def resolver(x):
        ...     global counter
        ...     counter += 1
        ...     return TABLE.get(x)
        >>> gen = xrange(10)
        >>> list1 = LazyResolvingList(gen, resolver)

    Any attempt to access the list should resolve the elements on demand,
    we will check the counter to ensure that we only resolve the elements
    that are requested::

        >>> counter
        0
        >>> list1[1]
        'b'
        >>> counter
        1
        >>> slice1 = list1[2:5]
        >>> isinstance(slice1, LazyResolvingList)
        True
        >>> counter
        1
        >>> list(slice1)
        ['c', 'd', 'e']
        >>> counter
        4
        >>> iter(list1).next()
        'a'
        >>> counter
        5
        >>> len(list1)
        10
        >>> counter
        5
        >>> list1[-1]
        'j'
        >>> counter
        6

    We can also add these lists together::

        >>> def gen2_func():
        ...     for i in xrange(10):
        ...         yield i
        >>> gen2 = gen2_func()
        >>> list2 = LazyResolvingList(gen, resolver)
        >>> list3 = LazyResolvingList(gen2, resolver)
        >>> list4 = list3 + list2
        >>> counter
        6
        >>> slice2 = list4[9:11]
        >>> counter
        6
        >>> isinstance(slice2, LazyResolvingList)
        True
        >>> list(slice2)
        ['j', 'a']
        >>> counter
        8
        >>> len(list4)
        20

    There are also some dictionary like methods which can be used to
    get at the tokens themselves::

        >>> list(list1.keys())
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        >>> list(list1.values())
        ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']
        >>> list(list1.items()) == zip(list(list1.keys()), list(list1.values()))
        True

    """
    __allow_access_to_unprotected_subobjects__ = True
    _util_method = None

    def __init__(self, initlist, resolver=None):
        self._data = LazyList(initlist)
        self._resolver = resolver

    def _resolve(self, token):
        """default resolver"""
        if self._resolver is not None:
            return self._resolver(token)
        if self._util_method is None:
            self._util_method = getUtility(IIntIds).getObject
        return self._util_method(token)

    def __len__(self):
        """This unspools the whole list""" 
        return len(self._data)

    def __iter__(self):
        """resolves the tokens"""
        for item in self._data:
            yield self._resolve(item)

    def __getitem__(self, i):
        return self._resolve(self._data[i])

    def __getslice__(self, i, j):
        """Returns an instance of this class containing the desired slice"""
        return self.__class__(self._data[i:j], self._resolver)

    def __add__(self, other):
        """Dumb add method that assumes identical resolvers"""
        assert self._resolver == getattr(other, '_resolver', None), \
                   "Both items must use the same resolver method"
        return self.__class__(self._data + other._data, self._resolver)

    def keys(self):
        return iter(self._data)

    def values(self):
        return iter(self)

    def items(self):
        for item in self._data:
            yield (item, self._resolve(item))

    def __nonzero__(self):
        return bool(self._data)


def lazyresolver(resolver=None, resolver_name=None):
    def wrapper(func, _resolver=resolver, _resolver_name=resolver_name):
        def resolver_func(self, *args, **kw):
            resolver = _resolver
            if not resolver and _resolver_name is not None:
                resolver = getattr(self, _resolver_name)
            return LazyResolvingList(func(self, *args, **kw),
                                     resolver = resolver)
        return resolver_func
    return wrapper

intidresolver = lazyresolver()



import unittest
from zope.app.testing import placelesssetup
from zope.testing.doctest import DocTestSuite
from Testing import ZopeTestCase as ztc
from Products.PloneTestCase import PloneTestCase as ptc
from OFS.SimpleItem import SimpleItem

from five.intid.lsm import USE_LSM
from five.intid.site import add_intids
from plone.app.relations.utils import add_relations

from Products.Five import zcml


class Demo(SimpleItem):
    def __init__(self, id):
        self.id = id
    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.id)

def contentSetUp(app):
    for i in range(30):
        oid = 'ob%d' % i
        app._setObject(oid, Demo(oid))

def base_setup(app):
    """Setup without basic CA stuff because PTC already provides this"""
    from plone.app import relations
    try:
        zcml.load_config('configure.zcml', relations)
    except:
        # XXX: When ptc has setup plone, then the zcml product
        # registration causes errors when run twice :-(
        pass
    if not USE_LSM:
        # monkey in our hooks
        from Products.Five.site.metaconfigure import classSiteHook
        from Products.Five.site.localsite import FiveSite
        from zope.interface import classImplements
        from zope.app.component.interfaces import IPossibleSite
        klass = app.__class__
        classSiteHook(klass, FiveSite)
        classImplements(klass, IPossibleSite)
    add_intids(app)
    add_relations(app)
    contentSetUp(app)

def setUp(app):
    """A setup that includes the basic CA setup"""
    placelesssetup.setUp()
    import Products.Five
    zcml.load_config('meta.zcml', Products.Five)
    zcml.load_config('permissions.zcml', Products.Five)
    zcml.load_config('configure.zcml', Products.Five)
    base_setup(app)


def tearDown():
    placelesssetup.tearDown()

# XXX: This messes things up, where else could we call it, it's only needed
# for workflow
ptc.setupPloneSite()

def test_suite():
    from plone.app.relations import local_role
    pas = DocTestSuite(local_role, setUp=placelesssetup.setUp(),
                       tearDown=placelesssetup.tearDown())

    readme = ztc.FunctionalDocFileSuite('README.txt',
                                        package='plone.app.relations')

    workflow = ztc.ZopeDocTestSuite('plone.app.relations.workflow',
                                    test_class=ptc.FunctionalTestCase,)

    return unittest.TestSuite([workflow, readme, pas])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



import unittest

from zope.testing import doctest
from zope.testing.doctestunit import DocTestSuite
from zope.app.testing import placelesssetup
from Testing import ZopeTestCase as ztc

# these are used by setup
from five.intid.site import add_intids
from five.intid.lsm import USE_LSM
from OFS.SimpleItem import SimpleItem

class Demo(SimpleItem):
    def __init__(self, obid):
        self.id = obid
    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.id)

def contentSetUp(app):
    for i in range(30):
        obid = 'ob%d' % i
        ob_object = Demo(obid)
        app._setObject(obid, ob_object)

def ChinatownSetUp(app):
    characters = ['jake', 'evelyn', 'noah', 'hollis', 'katherine',
                  'the past', 'investigation']
    for c in characters:
        app._setObject(c, Demo(c))

def setUp(app):
    # turn on all needed zcml
    placelesssetup.setUp()
    import Products.Five
    from Products.Five import zcml
    from plone import relations
    zcml.load_config('meta.zcml', Products.Five)
    zcml.load_config('permissions.zcml', Products.Five)
    zcml.load_config('configure.zcml', Products.Five)
    zcml.load_config('configure.zcml', relations)
    # Make a site, turn on the local site hooks, add the five.intid utility
    from zope.app.component.hooks import setSite, setHooks
    if not USE_LSM:
        # monkey in our hooks
        from Products.Five.site.metaconfigure import classSiteHook
        from Products.Five.site.localsite import FiveSite
        from zope.interface import classImplements
        from zope.app.component.interfaces import IPossibleSite
        klass = app.__class__
        classSiteHook(klass, FiveSite)
        classImplements(klass, IPossibleSite)
    add_intids(app)
    setSite(app)
    setHooks()
    contentSetUp(app)

def tearDown():
    placelesssetup.tearDown()

optionflags = doctest.ELLIPSIS
def test_suite():
    # Zope 2 + Five Integration tests that use ZopeTestCase
    integration = ztc.FunctionalDocFileSuite('container.txt',
                                             optionflags=optionflags,
                                             package='plone.relations')
    readme = ztc.FunctionalDocFileSuite('README.txt',
                                        optionflags=optionflags,
                                        package='plone.relations')


    lazy = DocTestSuite('plone.relations.lazylist')

    return unittest.TestSuite((lazy, integration, readme))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
